@@ -1,18 +1,18 @@
 import sys
-import logging
+
+from werkzeug.utils import cached_property
+from werkzeug._compat import iteritems
+from werkzeug._compat import to_unicode, to_bytes
 
 import jam
-import jam.lang.langs as langs
-import jam.common as common
+from .common import consts, json_defaul_handler
 
 class AbortException(Exception):
     pass
 
-class DebugException(Exception):
-    pass
-
 class AbstractItem(object):
-    def __init__(self, owner, name='', caption='', visible = True, item_type_id=0, js_filename=''):
+    def __init__(self, task, owner, name='', caption='', visible = True, item_type_id=0, js_filename=''):
+        self.task = task
         self.owner = owner
         self.item_name = name
         self.items = []
@@ -25,21 +25,46 @@ class AbstractItem(object):
                 owner.items.append(self)
                 if not hasattr(owner, self.item_name):
                     setattr(owner, self.item_name, self)
-            self.task = owner.task
         self.item_caption = caption
         self.visible = visible
         self.item_type_id = item_type_id
+        if self != task:
+            self.log = task.log
         self._loader = TracebackLoader(self)
 
+    def task_locked(self):
+        try:
+            if self.task.ID:
+                if self.master:
+                    owner = self.master.owner
+                else:
+                    owner = self.owner
+                if owner:
+                    return self.task.app.task_locked()
+        except:
+            pass
+
+    def __setattr__(self, name, value):
+        if self.task_locked():
+            raise Exception(self.task.language('server_tree_immutable') + \
+                ' Item: "%s", Attribute: "%s"' % (self.item_name, name))
+        super(AbstractItem, self).__setattr__(name, value)
+
     @property
     def session(self):
-        if hasattr(jam.context, 'session'):
+        try:
             return jam.context.session
+        except:
+            pass
 
     @property
     def environ(self):
-        if hasattr(jam.context, 'environ'):
-            return jam.context.environ
+        return jam.context.environ
+
+    @property
+    def user_info(self):
+        if self.session:
+            return self.session['user_info']
 
     def find(self, name):
         for item in self.items:
@@ -59,7 +84,7 @@ class AbstractItem(object):
         for item in self.items:
             item.all(func)
 
-    def write_info(self, info):
+    def write_info(self, info, server):
         info['id'] = self.ID
         info['name'] = self.item_name
         info['caption'] = self.item_caption
@@ -75,21 +100,21 @@ class AbstractItem(object):
         self.item_type_id = info['type']
         self.js_filename = info['js_filename']
 
-    def get_info(self):
+    def get_info(self, server=False):
         result = {}
         result['items'] = []
-        self.write_info(result)
+        self.write_info(result, server)
         for item in self.items:
-            result['items'].append((item.item_type_id, item.get_info()))
+            result['items'].append((item.item_type_id, item.get_info(server)))
         return result
 
-    def get_child_class(self, item_type_id):
+    def get_child_class(self):
         pass
 
     def set_info(self, info):
         self.read_info(info)
         for item_type_id, item_info in info['items']:
-            child = self.get_child_class(item_type_id)(self)
+            child = self.get_child_class()(self.task, self, item_info['name'])
             child.item_type_id = item_type_id
             child.set_info(item_info)
 
@@ -100,7 +125,7 @@ class AbstractItem(object):
         self.bind_item()
         for item in self.items:
             item.bind_items()
-        self.item_type = common.ITEM_TYPES[self.item_type_id - 1]
+        self.item_type = consts.ITEM_TYPES[self.item_type_id - 1]
 
     def get_module_name(self):
         result = self.owner.get_module_name() + '.' + self.item_name
@@ -108,18 +133,18 @@ class AbstractItem(object):
 
     def store_handlers(self):
         result = {}
-        for key, value in self.__dict__.iteritems():
+        for key, value in iteritems(self.__dict__):
             if key[0:3] == 'on_':
                 result[key] = self.__dict__[key]
         return result
 
     def clear_handlers(self):
-        for key, value in self.__dict__.iteritems():
+        for key, value in iteritems(self.__dict__):
             if key[0:3] == 'on_':
                 self.__dict__[key] = None
 
     def load_handlers(self, handlers):
-        for key, value in handlers.iteritems():
+        for key, value in iteritems(handlers):
             self.__dict__[key] = handlers[key]
 
     def get_master_field(self, fields, master_field):
@@ -136,29 +161,68 @@ class AbstractItem(object):
     def load_code(self):
         return self.item_name
 
+    def check_operation(self, operation):
+        try:
+            app = self.task.app
+            if not consts.SAFE_MODE:
+                return True
+            elif self.task == app.admin:
+                if consts.SAFE_MODE:
+                    session = self.session
+                    if session and session['user_info']['admin']:
+                        return True
+                else:
+                    return True
+            else:
+                session = self.session
+                if session:
+                    role_id = session['user_info']['role_id']
+                    privileges = self.task.app.get_privileges(role_id)
+                    priv_dic = privileges.get(self.ID)
+                    if priv_dic:
+                        return priv_dic[operation]
+        except:
+            return False
+
+    def can_view(self):
+        return self.check_operation('can_view')
+
+
 class AbstrGroup(AbstractItem):
     pass
 
 
 class AbstrTask(AbstractItem):
     def __init__(self, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
-        AbstractItem.__init__(self, owner, name, caption, visible, item_type_id, js_filename)
+        AbstractItem.__init__(self, self, owner, name, caption, visible, item_type_id, js_filename)
         self.task = self
-        self.__language = None
-        self.item_type_id = common.TASK_TYPE
+        self.item_type_id = consts.TASK_TYPE
         self.history_item = None
+        self.lock_item = None
         self.log = None
 
-    def write_info(self, info):
-        super(AbstrTask, self).write_info(info)
+    def task_locked(self):
+        try:
+            if self.ID:
+                return self.app.task_locked()
+        except:
+            pass
+
+    def write_info(self, info, server):
+        super(AbstrTask, self).write_info(info, server)
         info['lookup_lists'] = self.lookup_lists
         if self.history_item:
             info['history_item'] = self.history_item.ID
+        if self.lock_item:
+            info['lock_item'] = self.lock_item.ID
 
     def set_info(self, info):
         super(AbstrTask, self).set_info(info)
         self.bind_items()
 
+    def get_child_class(self):
+        pass
+
     def item_by_name(self, item_name):
         for group in self.items:
             if group.item_name == item_name:
@@ -183,59 +247,38 @@ class AbstrTask(AbstractItem):
                 self.compile_item(item)
         for group in self.items:
             for item in group.items:
-                if group.item_type_id != common.REPORTS_TYPE:
+                if group.item_type_id != consts.REPORTS_TYPE:
                     for detail in item.details:
                         self.compile_item(detail)
 
-    def get_language(self):
-        return self.__language
-
-    def set_language(self, value):
-        self.__language = value
-        self.lang = langs.get_lang_dict(value)
-        common.SETTINGS['LANGUAGE'] = value
-
-    language = property (get_language, set_language)
-
-    def get_settings(self):
-        return common.SETTINGS
-
-    def init_locale(self):
-        import locale
-        result = {}
-        try:
-            locale.setlocale(locale.LC_ALL, '')
-            loc = locale.localeconv()
-            for setting in common.LOCALE_SETTINGS:
-                try:
-                    common.SETTINGS[setting] = loc['setting'.lower()]
-                except:
-                    common.SETTINGS[setting] = common.DEFAULT_SETTINGS[setting]
-        except:
-            pass
-        try:
-            common.SETTINGS['D_FMT'] = locale.nl_langinfo(locale.D_FMT)
-        except:
-            common.SETTINGS['D_FMT'] = '%Y-%m-%d'
-        common.SETTINGS['D_T_FMT'] = '%s %s' % (common.D_FMT, '%H:%M')
+    def language(self, key):
+        return consts.language(key)
 
 class AbstrItem(AbstractItem):
-    def __init__(self, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
-        AbstractItem.__init__(self, owner, name, caption, visible, item_type_id, js_filename)
-        if not hasattr(self.task, self.item_name):
-            setattr(self.task, self.item_name, self)
+    def __init__(self, task, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
+        AbstractItem.__init__(self, task, owner, name, caption, visible, item_type_id, js_filename)
+        if not isinstance(self, AbstrDetail):
+            if self.owner and not hasattr(self.task, self.item_name):
+                setattr(self.task, self.item_name, self)
 
-    def write_info(self, info):
-        super(AbstrItem, self).write_info(info)
+    def write_info(self, info, server):
+        super(AbstrItem, self).write_info(info, server)
         info['fields'] = self.field_defs
         info['filters'] = self.filter_defs
         info['reports'] = self.get_reports_info()
         info['default_order'] = self._order_by
         info['primary_key'] = self._primary_key
         info['deleted_flag'] = self._deleted_flag
+        info['virtual_table'] = self.virtual_table
         info['master_id'] = self._master_id
         info['master_rec_id'] = self._master_rec_id
         info['keep_history'] = self.keep_history
+        info['edit_lock'] = self.edit_lock
+        info['view_params'] = self._view_list
+        info['edit_params'] = self._edit_list
+        info['virtual_table'] = self.virtual_table
+        if server:
+            info['table_name'] = self.table_name
 
     def read_info(self, info):
         super(AbstrItem, self).read_info(info)
@@ -243,17 +286,35 @@ class AbstrItem(AbstractItem):
         self.create_filters(info['filters'])
         self.reports = info['reports']
         self._order_by = info['default_order']
+        self._primary_key = info['primary_key']
+        self._deleted_flag = info['deleted_flag']
+        self._virtual_table = info['virtual_table']
+        self._master_id = info['master_id']
+        self._master_rec_id = info['master_rec_id']
+        self._keep_history = info['keep_history']
+        self.edit_lock = info['edit_lock']
+        self._view_list = info['view_params']
+        self._edit_list = info['edit_params']
+        self.table_name = info['table_name']
 
     def bind_item(self):
         self.prepare_fields()
         self.prepare_filters()
-#        self.fields = list(self._fields)
+
+    def can_create(self):
+        return self.check_operation('can_create')
+
+    def can_edit(self):
+        return self.check_operation('can_edit')
+
+    def can_delete(self):
+        return self.check_operation('can_delete')
 
 
 class AbstrDetail(AbstrItem):
 
-    def write_info(self, info):
-        super(AbstrDetail, self).write_info(info)
+    def write_info(self, info, server):
+        super(AbstrDetail, self).write_info(info, server)
         info['prototype_ID'] = self.prototype.ID
 
     def read_info(self, info):
@@ -261,16 +322,22 @@ class AbstrDetail(AbstrItem):
         self.owner.details.append(self)
         if not hasattr(self.owner.details, self.item_name):
             setattr(self.owner.details, self.item_name, self)
+        self._prototype_id = info['prototype_ID']
 
+    def bind_item(self):
+        super(AbstrDetail, self).bind_item();
+        if hasattr(self, '_prototype_id'):
+            self.prototype = self.task.item_by_ID(self._prototype_id)
+            self.init_fields()
 
 class AbstrReport(AbstractItem):
-    def __init__(self, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
-        AbstractItem.__init__(self, owner, name, caption, visible, item_type_id, js_filename)
+    def __init__(self, task, owner, name, caption, visible = True, item_type_id=0, js_filename=''):
+        AbstractItem.__init__(self, task, owner, name, caption, visible, item_type_id, js_filename)
         if not hasattr(self.task, self.item_name):
             setattr(self.task, self.item_name, self)
 
-    def write_info(self, info):
-        super(AbstrReport, self).write_info(info)
+    def write_info(self, info, server):
+        super(AbstrReport, self).write_info(info, server)
         info['fields'] = self.param_defs
 
     def read_info(self, info):
